# -*- coding: utf-8 -*-
# ===============LICENSE_START=======================================================
# Acumos Apache-2.0
# ===================================================================================
# Copyright (C) 2017-2018 AT&T Intellectual Property & Tech Mahindra. All rights reserved.
# ===================================================================================
# This Acumos software file is distributed by AT&T and Tech Mahindra
# under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# This file is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ===============LICENSE_END=========================================================
"""
Provides metadata generation utilities
"""
import importlib
import sys
from types import ModuleType
from pkg_resources import get_distribution, DistributionNotFound
from os.path import basename, normpath, sep as pathsep

from acumos.exc import AcumosError
from acumos.modeling import _is_namedtuple


BUILTIN_MODULE_NAMES = set(sys.builtin_module_names)
PACKAGE_DIRS = {'site-packages', 'dist-packages'}
SCHEMA_VERSION = '0.6.0'
_SCHEMA = "acumos.schema.model:{}".format(SCHEMA_VERSION)
_MEDIA_TYPES = {str: 'text/plain', bytes: 'application/octet-stream', dict: 'application/json', 'protobuf': 'application/vnd.google.protobuf'}

# known package mappings because Python packaging is madness
_REQ_MAP = {
    'sklearn': 'scikit-learn',
}


class Options(object):
    '''
    A collection of options that users may wish to specify along with their Acumos model

    Parameters
    ----------
    create_microservice : bool, optional
        If True, instructs the Acumos platform to eagerly build the model microservice
    deploy : bool, optional
        If True, ask to deploy model using Jenkins runner (create_microservice should be also True)
    license : str, optional
        A license to include with the Acumos model. This parameter may either be a path to a license
        file, or a string containing the license content.
    '''
    __slots__ = ('create_microservice', 'deploy', 'license')

    def __init__(self, create_microservice=True, deploy=False, license=None):
        self.create_microservice = create_microservice or deploy
        self.deploy = deploy
        self.license = license


class Requirements(object):
    '''
    A collection of optional user-provided Python requirement metadata

    Parameters
    ----------
    reqs : Sequence[str], optional
        A sequence of pip-installable Python package names
    req_map : Dict[str, str] or Dict[module, str], optional
        A corrective mapping of Python modules to pip-installable package names. For example
        `req_map={'sklearn': 'scikit-learn'}` or `import sklearn; req_map={sklearn: 'scikit-learn'}`.
    packages : Sequence[str], optional
        A sequence of paths to Python packages (i.e. directories with an __init__.py). Provided Python
        packages will be copied along with your model and added to the PYTHONPATH at runtime.
    scripts : Sequence[str], optional
        A sequence of paths to Python scripts. A path can point to a Python script, or directory
        containing Python scripts. If a directory is provided, all Python scripts within the directory
        will be included. Provided Python scripts will be copied along with your model and added to the
        PYTHONPATH at runtime.
    '''

    __slots__ = ('reqs', 'req_map', 'packages', 'scripts')

    def __init__(self, reqs=None, req_map=None, packages=None, scripts=None):
        self.reqs = set() if reqs is None else _safe_set(reqs)
        self.req_map = dict() if req_map is None else req_map.copy()
        self.req_map.update(_REQ_MAP)
        self.packages = set() if packages is None else _safe_set(packages)
        self.scripts = set() if scripts is None else _safe_set(scripts)

    @property
    def package_names(self):
        return frozenset(basename(p.rstrip(pathsep)) for p in self.packages)


def _safe_set(input_):
    '''Safely returns a set from input'''
    return {input_, } if isinstance(input_, str) else set(input_)


def create_model_meta(model, name, requirements, encoding='protobuf'):
    '''Returns a model metadata dictionary'''
    return {'schema': _SCHEMA,
            'runtime': _create_runtime(requirements, encoding),
            'name': name,
            'methods': {name: {'input': {'name': f.input_type.__name__,
                                         'media_type': [_MEDIA_TYPES[encoding if _is_namedtuple(f.input_type) else f.input_type.__supertype__._raw_type]],
                                         'metadata': {} if _is_namedtuple(f.input_type) else f.input_type.__supertype__._metadata,
                                         'description': '' if _is_namedtuple(f.input_type) else f.input_type.__supertype__._doc},
                               'output': {'name': f.output_type.__name__,
                                          'media_type': [_MEDIA_TYPES[encoding if _is_namedtuple(f.output_type) else f.output_type.__supertype__._raw_type]],
                                          'metadata': {} if _is_namedtuple(f.output_type) else f.output_type.__supertype__._metadata,
                                          'description': '' if _is_namedtuple(f.output_type) else f.output_type.__supertype__._doc},
                               'description': f.description} for name, f in model.methods.items()}}


def _create_runtime(requirements, encoding='protobuf'):
    '''Returns a runtime dict'''
    reqs = _gather_requirements(requirements)
    return {'name': 'python',
            'version': '.'.join(map(str, sys.version_info[:3])),
            'dependencies': _create_dependencies(reqs)}


def _gather_requirements(requirements):
    '''Yields (name, version) tuples of required 3rd party Python packages'''
    for req_name in _filter_requirements(requirements):
        yield _get_distribution(req_name)


def _filter_requirements(requirements):
    '''Returns a set required 3rd party Python package names'''
    # first get all non-stdlib requirement names
    req_names = (n for n in map(_get_requirement_name, requirements.reqs) if not _in_stdlib(n))

    # then apply any user-provided requirement mappings
    req_map = {_get_requirement_name(k): v for k, v in requirements.req_map.items()}
    mapped_reqs = {req_map.get(r, r) for r in req_names}

    # finally remove user-provided custom package names, as they won't exist in pip
    filtered_reqs = mapped_reqs - requirements.package_names
    return filtered_reqs


def _get_requirement_name(req):
    '''Returns the str name of a requirement'''
    if isinstance(req, ModuleType):
        name = req.__name__
    elif isinstance(req, str):
        name = req
    else:
        raise AcumosError("Requirement {} is invalid; must be ModuleType or string".format(req))
    return name


def _get_distribution(req_name):
    '''Returns (name, version) tuple given a requirement'''
    try:
        return str(get_distribution(req_name).as_requirement()).split('==')
    except DistributionNotFound:
        raise AcumosError("Module {} was detected as a dependency, but not found as a pip installed package. Use acumos.session.Requirements to declare custom packages or map module names to pip-installable names (e.g. Requirements(req_map=dict(PIL='pillow')) )".format(req_name))


def _in_stdlib(module_name):
    '''Returns True if the package name is part of the standard library (not airtight)'''
    if module_name in BUILTIN_MODULE_NAMES:
        return True

    try:
        install_path = importlib.import_module(module_name).__file__
    except ImportError:
        return False

    if not install_path.startswith(normpath(sys.base_prefix)):
        return False

    dirs = set(normpath(install_path).split(pathsep))
    return not PACKAGE_DIRS & dirs


def _create_dependencies(req_tuples):
    '''Returns a dict containing model dependency metadata'''
    return {
        'pip': {
            'indexes': [],
            'requirements': [{'name': n, 'version': v} for n, v in req_tuples]
        },
        'conda': {
            'channels': [],
            'requirements': []
        }
    }



# -*- coding: utf-8 -*-
# ===============LICENSE_START=======================================================
# Acumos Apache-2.0
# ===================================================================================
# Copyright (C) 2017-2018 AT&T Intellectual Property & Tech Mahindra. All rights reserved.
# ===================================================================================
# This Acumos software file is distributed by AT&T and Tech Mahindra
# under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# This file is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ===============LICENSE_END=========================================================
"""
Provides a Acumos session for pushing and dumping models
"""
import os
import random
import string
import shutil
import json
from typing import Optional, Union

import requests
import fnmatch
import warnings
from contextlib import contextmanager, ExitStack
from tempfile import TemporaryDirectory
from os import walk, mkdir
from os.path import extsep, exists, abspath, dirname, isdir, isfile, expanduser, relpath, basename, join as path_join
from pathlib import Path
from collections import namedtuple
from glob import glob

from acumos.pickler import AcumosContextManager, dump_model
from acumos.metadata import create_model_meta, Requirements, Options
from acumos.utils import dump_artifact, get_qualname
from acumos.exc import AcumosError
from acumos.protogen import model2proto, compile_protostr
from acumos.logging import get_logger
from acumos.modeling import Model
from acumos.auth import get_jwt, clear_jwt


logger = get_logger(__name__)

_LICENSE_NAME = 'license.json'
_PYEXT = "{}py".format(extsep)
_PYGLOB = "*{}".format(_PYEXT)

_ServerResponse = namedtuple('ServerResponse', 'status_code reason text')
_DEPR_MSG = ('Usage of `auth_api` is deprecated; provide an onboarding token instead. '
             'See https://pypi.org/project/acumos/ for more information.')


class AcumosSession(object):
    '''
    A session that enables onboarding models to Acumos

    Parameters
    ----------
    push_api : str
        The full URL to the Acumos onboarding server upload API
    auth_api : str
        The full URL to the Acumos onboarding server authentication API.

        .. deprecated:: 0.7.1
            Users should provide an onboarding token instead of username and password credentials.
    '''

    def __init__(self, push_api=None, auth_api=None):
        self.push_api = push_api
        self.auth_api = auth_api

        if auth_api is not None:
            warnings.warn(_DEPR_MSG, DeprecationWarning, stacklevel=2)

    def push(self, model, name, requirements=None, extra_headers=None, options=None) -> Optional[str]:
        '''
        Pushes a model to Acumos, returns the docker image URI.

        Parameters
        ----------
        model : ``acumos.modeling.Model``
            An Acumos model instance
        name : str
            The name of your model
        requirements : ``acumos.metadata.Requirements``, optional
            Additional Python dependencies that you can optionally specify
        extra_headers : dict, optional
            Additonal HTTP headers included in the POST to the Acumos onboarding server
        options : ``acumos.metadata.Options``, optional
            Additional Acumos options that you can optionally specify
        '''
        _assert_valid_input(model, requirements)
        _assert_valid_api('push_api', self.push_api, True)
        _assert_valid_api('auth_api', self.auth_api, False)
        options = _validate_options(options)

        with _dump_model(model, name, requirements) as dump_dir:
            return _push_model(dump_dir, self.push_api, self.auth_api, options, extra_headers=extra_headers)

    def dump(self, model: Model, name: str, outdir: Union[Path, str], requirements: Optional[Requirements] = None, replace: bool = False):
        '''
        Creates a directory located at ``outdir/name`` containing Acumos model artifacts

        Parameters
        ----------
        model : ``acumos.modeling.Model``
            An Acumos model instance
        name : str
            The name of your model
        outdir : str or Path
            The directory or folder to save your model .zip to
        requirements : ``acumos.metadata.Requirements``, optional
            Additional Python dependencies that you can optionally specify
        replace: bool
            If the model dir already exists, acumos will fail unless replace is set to True
        '''
        _assert_valid_input(model, requirements)

        outdir = Path(expanduser(outdir))

        if outdir.exists() and replace:
            shutil.rmtree(outdir)

        with _dump_model(model, name, requirements) as dump_dir:
            _copy_dir(dump_dir, outdir, name)

    def dump_zip(self, model: Model, name: str, outfile: Union[Path, str], requirements: Optional[Requirements] = None, replace: bool = False):
        '''
        Creates a zipped package located at ``outfile``

        Parameters
        ----------
        model : ``acumos.modeling.Model``
            An Acumos model instance
        name : str
            The name of your model
        outfile : str or Path
            The name or path to the ouput .zip
        requirements : ``acumos.metadata.Requirements``, optional
            Additional Python dependencies that you can optionally specify
        replace: bool
            If the model zip already exists, acumos will fail unless replace is set to True
        '''
        _assert_valid_input(model, requirements)
        outfile = Path(outfile)
        if not outfile.name.lower().endswith(".zip"):
            raise AcumosError("outfile must have a zip extension.")
        with _dump_model(model, name, requirements) as dump_dir:
            import zipfile
            if outfile.exists() and not replace:
                raise AcumosError("Model {} has already been dumped, set replace to True to overwrite.".format(outfile))
            with zipfile.ZipFile(outfile, 'w') as model_zip:
                for root, _, files in os.walk(dump_dir):
                    for file in files:
                        model_zip.write(filename=os.path.join(root, file), arcname=file)


def _validate_options(options):
    '''Validates and returns an `Options` object'''
    if options is None:
        options = Options()
    elif not isinstance(options, Options):
        raise AcumosError('The `options` parameter must be of type `acumos.metadata.Options`')
    return options


def _assert_valid_input(model, requirements):
    '''Raises AcumosError if inputs are invalid'''
    if not isinstance(model, Model):
        raise AcumosError("Input `model` must be of type {}".format(get_qualname(Model)))

    if requirements is not None and not isinstance(requirements, Requirements):
        raise AcumosError("Input `requirements` must be of type {}".format(get_qualname(Requirements)))


def _assert_valid_api(param, api, required):
    '''Raises AcumosError if an api is invalid'''
    if api is None:
        if required:
            raise AcumosError("AcumosSession.push requires that the API for `{}` be provided".format(param))
    else:
        if not api.startswith('https'):
            logger.warning("Provided `{}` API {} does not begin with 'https'. Your password and token are visible in plaintext!".format(param, api))


def _push_model(dump_dir, push_api, auth_api, options, max_tries=2, extra_headers=None) -> Optional[str]:
    '''Pushes a model to the Acumos server, returns the docker image URI if requested'''
    with ExitStack() as stack:
        model = stack.enter_context(open(path_join(dump_dir, 'model.zip'), 'rb'))
        meta = stack.enter_context(open(path_join(dump_dir, 'metadata.json')))
        proto = stack.enter_context(open(path_join(dump_dir, 'model.proto')))

        files = {'model': ('model.zip', model, 'application/zip'),
                 'metadata': ('metadata.json', meta, 'application/json'),
                 'schema': ('model.proto', proto, 'text/plain')}

        # include a license if one is provided
        if options.license is not None:
            _add_license(dump_dir, options.license)
            license = stack.enter_context(open(path_join(dump_dir, _LICENSE_NAME)))
            files['license'] = (_LICENSE_NAME, license, 'application/json')

        tries = 1
        return _post_model(files, push_api, auth_api, tries, max_tries, extra_headers, options)


def _add_license(rootdir, license_str):
    '''Adds a license file to the model root directory'''
    license_dst = path_join(rootdir, _LICENSE_NAME)
    if isfile(license_str):
        shutil.copy(license_str, license_dst)
    else:
        license_dict = {'license': license_str}  # the license team hasn't defined a license schema yet
        dump_artifact(license_dst, data=license_dict, module=json, mode='w')


def _post_model(files, push_api, auth_api, tries, max_tries, extra_headers, options) -> Optional[str]:
    '''Attempts to post the model to Acumos, returns the docker image URI'''
    headers = {'Authorization': get_jwt(auth_api),
               'isCreateMicroservice': 'true' if options.create_microservice else 'false',
               'deploy': 'true' if options.deploy else 'false'}

    if extra_headers is not None:
        headers.update(extra_headers)

    resp = requests.post(push_api, files=files, headers=headers)
    if resp.ok:
        logger.info("Model pushed successfully to {}".format(push_api))
        if options.create_microservice:
            try:
                docker_image_uri = resp.json()["dockerImageUri"]
                logger.info(f"Acumos model docker image successfully created: {docker_image_uri}")
                return docker_image_uri
            except KeyError:
                logger.warning("Docker image URI could not be found in server response, "
                               "on-boarding server is probably running a version prior to Demeter.")
    else:
        clear_jwt()
        if resp.status_code == 401 and tries != max_tries:
            logger.warning('Model push failed due to an authorization failure. Clearing credentials and trying again')
            _post_model(files, push_api, auth_api, tries + 1, max_tries, extra_headers, options)
        else:
            raise AcumosError("Model push failed: {}".format(_ServerResponse(resp.status_code, resp.reason, resp.text)))


@contextmanager
def _dump_model(model, name, requirements=None):
    '''Generates model artifacts and serializes the model'''
    requirements = Requirements() if requirements is None else requirements

    with TemporaryDirectory() as rootdir:

        model_dir = path_join(rootdir, 'model')
        mkdir(model_dir)

        with AcumosContextManager(model_dir) as context:

            with open(context.build_path('model.pkl'), 'wb') as f:
                dump_model(model, f)

            # generate protobuf definition
            proto_pkg = context.parameters['protobuf_package'] = _random_string()
            protostr = model2proto(model, proto_pkg)
            dump_artifact(rootdir, 'model.proto', data=protostr, module=None, mode='w')

            # generate protobuf source code
            module_name = 'model'
            proto_dir = context.create_subdir('scripts', 'acumos_gen', proto_pkg)
            compile_protostr(protostr, proto_pkg, module_name, proto_dir)

            # generate model metadata
            requirements.reqs.update(context.package_names)
            metadata = create_model_meta(model, name, requirements)
            dump_artifact(rootdir, 'metadata.json', data=metadata, module=json, mode='w')

            # bundle user-provided code
            code_dir = context.create_subdir('scripts', 'user_provided')
            Path(code_dir, '.keep').touch()  # may resolve pruning issues when unzipping

            # copy packages and modules
            pkg_scripts = _gather_package_scripts(requirements.packages)
            _copy_package_scripts(context, pkg_scripts, code_dir)

            scripts = set(_gather_scripts(context, requirements))
            _copy_scripts(scripts, code_dir)

        shutil.make_archive(model_dir, 'zip', model_dir)  # create zip at same level as parent
        shutil.rmtree(model_dir)  # clean up model directory

        yield rootdir


def _copy_dir(src_dir, outdir, name):
    '''Copies a directory to a new location'''
    dst_path = path_join(outdir, name)
    if isdir(dst_path):
        raise AcumosError("Model {} has already been dumped".format(dst_path))
    shutil.copytree(src_dir, dst_path)


def _copy_package_scripts(context, scripts, code_dir):
    '''Moves all gathered package scripts to the context code directory'''
    for script_relpath, script_abspath in scripts:
        script_dirname = dirname(script_relpath)
        context_absdir = context.create_subdir(code_dir, script_dirname, exist_ok=True)
        shutil.copy(script_abspath, context_absdir)


def _gather_package_scripts(packages):
    '''Yields (relpath, abspath) tuples of Python scripts from a sequence of packages'''
    for path in packages:
        path = expanduser(path)
        if not isdir(path):
            raise AcumosError("Path {} is not a directory".format(path))

        for root, dirnames, filenames in walk(path):
            for filename in fnmatch.filter(filenames, '*.py'):
                script_abspath = path_join(root, filename)
                script_relpath = path_join(basename(path), relpath(script_abspath, path))
                yield script_relpath, script_abspath


def _gather_scripts(context, reqs):
    '''Yields absolute paths of Python script dependencies'''
    for script in context.scripts:
        # script can be a python built-in module with no associated file
        if script.__spec__.origin != 'built-in':
            yield script.__file__

    for script_path in reqs.scripts:
        script_abspath = abspath(expanduser(script_path))

        if not exists(script_abspath):
            raise AcumosError("Provided script requirement {} does not exist".format(script_path))

        if isdir(script_abspath):
            globbed_scripts = glob(path_join(script_abspath, _PYGLOB))
            if not globbed_scripts:
                raise AcumosError("Provided script requirement directory {} does not contain Python scripts".format(script_path))
            else:
                yield from globbed_scripts
        elif isfile(script_abspath) and script_abspath.endswith(_PYEXT):
            yield script_abspath
        else:
            raise AcumosError("Provided script requirement {} is invalid. See acumos.metadata.Requirements for documentation".format(script_path))


def _copy_scripts(script_paths, code_dir):
    '''Copies individual scripts to a user-provided code directory'''
    for path in script_paths:
        shutil.copy(path, code_dir)


def _random_string(length=32):
    '''Returns a random string containing ascii characters'''
    return ''.join(random.choice(string.ascii_letters) for _ in range(length))



# -*- coding: utf-8 -*-
# ===============LICENSE_START=======================================================
# Acumos Apache-2.0
# ===================================================================================
# Copyright (C) 2017-2018 AT&T Intellectual Property & Tech Mahindra. All rights reserved.
# ===================================================================================
# This Acumos software file is distributed by AT&T and Tech Mahindra
# under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# This file is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ===============LICENSE_END=========================================================
"""
Provides custom pickle utilities
"""
import sys
import json
import inspect
import contextlib
import tempfile
from os import makedirs
from os.path import basename, isdir, isfile, join as path_join
from copy import deepcopy
from functools import partial
from typing import Dict, List
from typing_inspect import NEW_TYPING
from types import ModuleType
from importlib import import_module

import dill

from acumos.modeling import _is_namedtuple, create_namedtuple, Empty
from acumos.exc import AcumosError
from acumos.utils import namedtuple_field_types


_DEFAULT_MODULES = ('acumos', 'dill')
_BLACKLIST = {'builtins', }
_DEFAULT = 'default'

_contexts = dict()

dump_model = partial(dill.dump, recurse=True)
dumps_model = partial(dill.dumps, recurse=True)
load_model = dill.load
loads_model = dill.loads


if not NEW_TYPING:
    from typing import GenericMeta

    def _save_annotation(pickler, obj):
        '''Workaround for dill annotation serialization bug'''
        if obj.__origin__ in (Dict, List):
            # recursively save object
            t = obj.__origin__
            args = obj.__args__
            pickler.save_reduce(_load_annotation, (t, args), obj=obj)
        else:
            # eventually hit base type, then use stock pickling logic. temp revert prevents infinite recursion
            t = obj
            args = None
            with _revert_dispatch(GenericMeta):
                pickler.save_reduce(_load_annotation, (t, args), obj=obj)

    def _load_annotation(t, args):
        '''Workaround for dill annotation serialization bug'''
        if t is Dict and args is not None:
            return Dict[args[0], args[1]]
        elif t is List and args is not None:
            return List[args[0]]
        else:
            return t

    dill.Pickler.dispatch[GenericMeta] = _save_annotation


def _save_namedtuple(pickler, obj):
    '''Workaround for dill NamedTuple serialization bug'''
    field_types = namedtuple_field_types(obj)
    pickler.save_reduce(_load_namedtuple, (obj.__name__, field_types), obj=obj)


def _load_namedtuple(name, field_types):
    '''Workaround for dill NamedTuple serialization bug'''
    return create_namedtuple(name, [(k, v) for k, v in field_types.items()])


@contextlib.contextmanager
def _revert_dispatch(t):
    '''Temporarily removes a type from the dispatch table to use existing serialization logic'''
    f = dill.Pickler.dispatch.pop(t)
    try:
        yield
    finally:
        dill.Pickler.dispatch[t] = f


def _save_keras(pickler, obj):
    '''Serializes a keras model to a context directory'''
    base_mod_name = _get_base_module(obj).__name__
    if base_mod_name == 'keras':
        import keras
    else:
        from tensorflow import keras

    backend = keras.backend.backend
    load_func = keras.models.load_model
    model_cls = keras.models.Model

    context = get_context()
    model_subdir = context.create_subdir()
    model_abspath, model_relpath = _add_file(model_subdir, 'model.h5')

    obj.save(model_abspath)  # /path/to/context/root/abc123/model.h5
    context.add_module('h5py')  # needed for keras model serialization
    context.add_module(backend())  # adds name of active keras backend

    # check for custom or contrib layer modules
    custom_objects = {}
    for layer in _get_keras_layers(obj, model_cls):
        module = _get_base_module(layer)
        if module.__name__ != base_mod_name:
            context.add_module(module)
            layer_cls = layer.__class__
            custom_objects[layer_cls.__name__] = layer_cls

    # store subpath because context root can change, and special layer classes to have them imported before load
    pickler.save_reduce(_load_keras, (model_relpath, custom_objects, load_func), obj=obj)


def _get_keras_layers(model, model_cls):
    '''Recursively walks a keras model and returns its layers'''
    for layer in model.layers:
        if isinstance(layer, model_cls):
            yield from _get_keras_layers(layer, model_cls)
        else:
            yield layer


def _load_keras(model_relpath, custom_objects, load_func):
    '''Loads a keras model from a context subdirectory'''
    context = get_context()
    model_path = context.build_path(*model_relpath)  # /different/path/to/context/root/abc123/model.h5
    model = load_func(model_path, custom_objects=custom_objects)
    return model


def _save_tf_tensor(pickler, tensor):
    '''Saves a TensorFlow tensor object'''
    import tensorflow as tf
    pickler.save_reduce(_load_tf_tensor, (tensor.name, tensor.graph, tf.get_default_session()), obj=tensor)


def _load_tf_tensor(name, graph, session):
    '''Loads a TensorFlow tensor object'''
    return graph.get_tensor_by_name(name)


def _save_tf_operation(pickler, op):
    '''Saves a TensorFlow operation object'''
    import tensorflow as tf
    pickler.save_reduce(_load_tf_operation, (op.name, op.graph, tf.get_default_session()), obj=op)


def _load_tf_operation(name, graph, session):
    '''Loads a TensorFlow operation object'''
    return graph.get_operation_by_name(name)


def _save_tf_session(pickler, session):
    '''Saves a TensorFlow session'''
    import tensorflow as tf

    context = get_context()
    model_subdir = context.create_subdir()
    model_abspath, model_relpath = _add_file(model_subdir, 'model')

    graph = session.graph
    with graph.as_default():
        saver = tf.train.Saver(allow_empty=True)
        saver.save(session, model_abspath, write_meta_graph=False)  # don't export meta graph twice

    pickler.save_reduce(_load_tf_session, (model_relpath, session.__class__, graph), obj=session)


def _load_tf_session(model_subpath, session_cls, graph):
    '''Loads a TensorFlow session'''
    import tensorflow as tf

    context = get_context()
    model_path = context.build_path(*model_subpath)

    with graph.as_default():
        sess = session_cls()
        saver = tf.train.Saver(allow_empty=True)
        saver.restore(sess, model_path)
    return sess


def _save_tf_graph(pickler, graph):
    '''Saves a TensorFlow graph'''
    import tensorflow as tf

    context = get_context()
    model_subdir = context.create_subdir()
    graph_abspath, graph_relpath = _add_file(model_subdir, 'graph.meta')

    with graph.as_default():
        tf.train.export_meta_graph(graph_abspath)
    pickler.save_reduce(_load_tf_graph, (graph_relpath, ), obj=graph)


def _load_tf_graph(graph_relpath):
    '''Loads a TensorFlow graph'''
    import tensorflow as tf

    context = get_context()
    graph_abspath = context.build_path(*graph_relpath)

    graph = tf.Graph()
    with graph.as_default():
        tf.train.import_meta_graph(graph_abspath)
    return graph


def _add_file(subdir, name):
    '''Helper function which returns the absolute and context-relative path of a file to be added'''
    file_abspath = path_join(subdir, name)
    file_relpath = (basename(subdir), name)
    return file_abspath, file_relpath


_CUSTOM_DISPATCH = {
    'keras.engine.training.Model': _save_keras,
    'tensorflow.python.framework.ops.Tensor': _save_tf_tensor,
    'tensorflow.python.framework.ops.Operation': _save_tf_operation,
    'tensorflow.python.client.session.BaseSession': _save_tf_session,
    'tensorflow.python.framework.ops.Graph': _save_tf_graph,
    'tensorflow.python.keras.engine.training.Model': _save_keras,
}


@contextlib.contextmanager
def _patch_dill():
    '''Temporarily patches the dill Pickler dispatch table to support custom serialization within a context'''
    try:
        dispatch = dill.Pickler.dispatch
        dill.Pickler.dispatch = deepcopy(dispatch)

        pickler_save = dill.Pickler.save

        def wrapped_save(pickler, obj, save_persistent_id=True):
            '''Hook that intercepts objects about to be saved'''
            _catch_object(obj)

            if _is_namedtuple(obj) and obj is not Empty:
                _save_namedtuple(pickler, obj)
            else:
                pickler_save(pickler, obj, save_persistent_id)

        dill.Pickler.save = wrapped_save
        yield
    finally:
        dill.Pickler.dispatch = dispatch
        dill.Pickler.save = pickler_save


def _catch_object(obj):
    '''Inspects object and executes custom serialization / bookkeeping logic'''

    # dynamically extend dispatch table to prevent unnecessary imports / dependencies
    obj_type = obj if inspect.isclass(obj) else type(obj)
    if obj_type not in dill.Pickler.dispatch:
        for path in _get_mro_paths(obj_type):
            if path in _CUSTOM_DISPATCH:
                dill.Pickler.dispatch[obj_type] = _CUSTOM_DISPATCH[path]

    base_module = _get_base_module(obj)
    if base_module is not None and base_module.__name__ not in _BLACKLIST:
        context = get_context()
        context.add_module(base_module)


def _get_base_module(obj):
    '''Returns the base module for a given object'''
    module = inspect.getmodule(obj)
    if module is not None:
        base_name, _, _ = module.__name__.partition('.')
        base_module = sys.modules[base_name]
    else:
        base_module = None
    return base_module


def _get_mro_paths(type_):
    '''Yields import path string for each entry in `inspect.getmro`'''
    for t in inspect.getmro(type_):
        yield "{}.{}".format(t.__module__, t.__name__)


class AcumosContext(object):
    '''Represents a workspace for a model that is being dumped'''

    def __init__(self, root_dir):
        if not isdir(root_dir):
            raise AcumosError("AcumosContext root directory {} does not exist".format(root_dir))
        self._modules = set()
        self._root_dir = root_dir
        self._params_path = path_join(root_dir, 'context.json')
        self.parameters = self._load_params()

        for mod in _DEFAULT_MODULES:
            self.add_module(mod)

    def create_subdir(self, *paths, exist_ok=False):
        '''Creates a new directory within the context root and returns the absolute path'''
        if not paths:
            tdir = tempfile.mkdtemp(dir=self._root_dir)
        else:
            tdir = path_join(self._root_dir, *paths)
            makedirs(tdir, exist_ok=exist_ok)
        return tdir

    def build_path(self, *paths):
        '''Returns an absolute path starting from the context root'''
        return path_join(self._root_dir, *paths)

    def add_module(self, module):
        '''Adds a module to the context module set'''
        if isinstance(module, str):
            try:
                module = import_module(module)
            except ImportError:
                raise AcumosError("Module '{}' was identified as a dependency, but cannot be imported. Ensure that it is installed and available".format(module))
        elif not isinstance(module, ModuleType):
            raise AcumosError("Module must be of type str or types.ModuleType, not {}".format(type(module)))

        self._modules.add(module)

    @property
    def abspath(self):
        '''Absolute path of the context root directory'''
        return self._root_dir

    @property
    def basename(self):
        '''Base name of the context root directory'''
        return basename(self._root_dir)

    @property
    def modules(self):
        '''The set of all modules (i.e. typing.ModuleType) identified as dependencies'''
        return frozenset(self._modules)

    @property
    def packages(self):
        '''The set of all base packages (i.e. typing.ModuleType with a package) identified as dependencies'''
        return frozenset(mod for mod in self._modules if mod.__package__)

    @property
    def package_names(self):
        '''The set of all base package names identified as dependencies'''
        return frozenset(mod.__name__ for mod in self.packages)

    @property
    def scripts(self):
        '''The set of all scripts (i.e. typing.ModuleType with no package) identified as dependencies'''
        return frozenset(script for script in (mod for mod in self._modules if not mod.__package__) if script.__name__ != '__main__')

    @property
    def script_names(self):
        '''The set of all script names identified as dependencies'''
        return frozenset(mod.__name__ for mod in self.scripts)

    def _load_params(self):
        '''Returns a parameters dict'''
        if isfile(self._params_path):
            with open(self._params_path) as f:
                return json.load(f)
        else:
            return dict()

    def save_params(self):
        '''Saves a parameters json file within the context root'''
        with open(self._params_path, 'w') as f:
            json.dump(self.parameters, f)


@contextlib.contextmanager
def AcumosContextManager(rootdir=None, name=_DEFAULT):
    '''Context manager that provides a AcumosContext object'''
    with _patch_dill():
        if name in _contexts:
            raise AcumosError("AcumosContext '{}' has already been created. Use `get_context` to access it.".format(name))
        try:
            with _DirManager(rootdir) as rootdir:
                context = AcumosContext(rootdir)
                _contexts[name] = context
                yield context
                context.save_params()
        finally:
            del _contexts[name]


@contextlib.contextmanager
def _DirManager(dir_=None):
    '''Wrapper that passes dir_ through or creates a temporary directory'''
    if dir_ is not None:
        if not isdir(dir_):
            raise AcumosError("Provided AcumosContext rootdir {} does not exist".format(dir_))
        yield dir_
    else:
        with tempfile.TemporaryDirectory() as tdir:
            yield tdir


def get_context(name=_DEFAULT):
    '''Returns an existing AcumosContext'''
    if name in _contexts:
        return _contexts[name]
    else:
        raise AcumosError("AcumosContext '{}' has not been created".format(name))



# -*- coding: utf-8 -*-
# ===============LICENSE_START=======================================================
# Acumos Apache-2.0
# ===================================================================================
# Copyright (C) 2017-2018 AT&T Intellectual Property & Tech Mahindra. All rights reserved.
# ===================================================================================
# This Acumos software file is distributed by AT&T and Tech Mahindra
# under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# This file is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ===============LICENSE_END=========================================================
"""
Provides model wrapping utilities
"""
from importlib.util import spec_from_file_location, module_from_spec
import os
import inspect
import contextlib
from collections import OrderedDict, namedtuple
import typing_inspect

from acumos.exc import AcumosError


def namedtuple_field_types(nt):
    '''Returns an OrderedDict corresponding to NamedTuple field types'''
    field_types = nt._field_types
    return OrderedDict((field, field_types[field]) for field in nt._fields)


def load_module(fullname, path):
    '''Imports and returns a module from path for Python 3.5+'''
    spec = spec_from_file_location(fullname, path)
    module = module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


def load_artifact(*path, module, mode):
    '''Artifact loader helper'''
    with open(os.path.join(*path), mode) as f:
        return module.load(f)


def dump_artifact(*path, data, module, mode):
    '''Artifact saver helper'''
    with open(os.path.join(*path), mode) as f:
        if module is None:
            f.write(data)
        else:
            module.dump(data, f)


def get_qualname(o):
    if inspect.isclass(o):
        return "{}.{}".format(o.__module__, o.__name__)
    else:
        return get_qualname(o.__class__)


@contextlib.contextmanager
def reraise(prefix, prefix_args):
    '''Reraises an exception with a more informative prefix'''
    try:
        yield
    except AcumosError as e:
        raise AcumosError("{}: {}".format(prefix.format(*prefix_args), e)).with_traceback(e.__traceback__)


InspectedType = namedtuple('InspectedType', ['type',    # The real type
                                             'origin',  # The unsubscribed / unaliased typed
                                             'args',    # The type arguments
                                             ])


def inspect_type(t) -> InspectedType:
    """Returns basic information on a type as an InspectedType"""
    args = typing_inspect.get_args(t)
    if typing_inspect.is_generic_type(t):
        origin = typing_inspect.get_origin(t)
    else:
        origin = t
    return InspectedType(type=t, origin=origin, args=args)
